#pragma once

void test_prepare_print_basic (void);
void test_prepare_print_delay (void);
void test_prepare_print_cancel (void);
void test_prepare_print_close (void);
void test_prepare_print_lockdown (void);
void test_prepare_print_results (void);
void test_prepare_print_parallel (void);

void test_print_basic (void);
void test_print_delay (void);
void test_print_cancel (void);
void test_print_close (void);
void test_print_lockdown (void);
void test_print_parallel (void);
